/*
**                    COWLOOP block device driver
** =======================================================================
** Read-write loop-driver with copy-on-write functionality.
**
** Two file-names should be supplied during insmod:
**
**           insmod cowloop rdofile=..... cowfile=....
**
**  - rdofile	read-only file(system)
**  - cowfile   storage-space for modified blocks of read-only file(system)
**
** The read-only file may be of type 'regular' or 'block-device'.
** The cow file is always a 'regular' file. If an existing regular file
** is used as cowfile, its contents will be used again for 
** the current read-only file.
** =======================================================================
** Author: Gerlof Langeveld - AT Computing 
** Email:  gerlof@ATComputing.nl
** Date:   March 2003
** (C) Copyright 2003, AT Consultancy, Nijmegen (The Netherlands)
**
** Modified:
**	20030908 - correct calculation of last bitmap chunk in case the
**	           entire bitmap is an exact multiple of MAPMAXSZ
**
** =======================================================================
** Inspired by
**    loop.c  by Theodore Ts'o and
**    cloop.c by Paul `Rusty' Russell & Klaus Knopper.
** =======================================================================
** Design-considerations:
**
**   The cowloop-driver starts a kernel-thread. All read- and write-request
**   on the read-only file and copy-on-write file are handled in the context
**   of that thread.
**   For the first experiments with the cowloop-driver, the request-queue
**   (INIT_REQUEST etc) made use of the do_generic_file_read() which worked
**   fine except in combination with the cloop-driver; that combination
**   resulted in a non-interruptible hangup of the system during
**   heavy load. Other experiments using the `make_request' interface also
**   resulted in unpredictable system hangups (even with proper use of
**   spin-locks).
**   Therefor a scheme has been designed to wakeup the kernel-thread as
**   soon as I/O-requests are available in the request-queue; this thread
**   handles the requests one-by-one by calling the proper read- or
**   write-function related to the open read-only file or copy-on-write file.
**   When all pending requests have been handled, the kernel-thread goes
**   back to sleep-state.
**   This approach requires some additional context-switches; however the
**   performance loss during heavy I/O is less than 3%.
**
** ---------------------------------------------------------------------------
** This program is free software; you can redistribute it and/or modify it
** under the terms of the GNU General Public License as published by the
** Free Software Foundation; either version 2, or (at your option) any
** later version.
**
** This program is distributed in the hope that it will be useful, but
** WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
** See the GNU General Public License for more details.
**
** $Log: cowloop.c,v $
** Revision 1.2  2004/06/30 14:13:33  gerlof
** Prepared for distribution.
**
** Revision 1.1  2004/06/30 08:38:40  gerlof
** Initial revision
**
*/

#include "kernelvers.h"
#include <linux/config.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/errno.h>
#include <linux/kernel.h>
#include <linux/major.h>
#include <linux/sched.h>
#include <linux/fs.h>
#include <linux/file.h>
#include <linux/stat.h>
#include <linux/vmalloc.h>
#include <linux/slab.h>
#include <asm/semaphore.h>
#include <asm/uaccess.h>
#include <linux/proc_fs.h>
#include "version.h"

EXPORT_NO_SYMBOLS;

#ifdef MODULE_LICENSE
MODULE_LICENSE("GPL");
#endif

#define	MAJOR_NR	COWMAJOR

#define DEVICE_NAME		"cowloop"
#define DEVICE_REQUEST		cowlo_request
#define DEVICE_NR(device)	(MINOR(device))
#define DEVICE_ON(device)
#define DEVICE_OFF(device)
#define DEVICE_NO_RANDOM
#include <linux/blk.h>

static char *rdofile;
static char *cowfile;

MODULE_PARM(rdofile, "s");
MODULE_PARM(cowfile, "s");

#define	MAPMAXNR	10	/* maximum number of bitmaps	            */
#define	MAPMAXSZ	65536	/* maximum bytes per bitmap (do not change) */

#define	CALCMAP(x)	((x)/(MAPMAXSZ*8))
#define	CALCBYTE(x)	(((x)%(MAPMAXSZ*8))>>3)
#define	CALCBIT(x)	((x)&7)

#define ALLCOW		1
#define ALLRDO		2
#define MIXEDUP		3


typedef	unsigned long int	pos_t;

static int	cowloop_hardsect = 512;
static int	cowloop_blocksz;
static int	cowloop_maxsect;

static int	cowlo_checkio (int,         loff_t);
static long int cowlo_readrdo (char *, int, loff_t);
static long int cowlo_readcow (char *, int, loff_t);
static long int cowlo_writecow(char *, int, loff_t);

struct cowloop_device
{
        /*
	** open file pointers
	*/
        struct file  	*rdofp,  *cowfp;	/* open file pointers      */

	/*
	** administration about read-only file
	*/
	int		numblocks;	/* number of block input file (Kb) */
	kdev_t		belowdev;	/* block device below us           */
	int		*undersz;	/* pointer to blksize_size of      */
					/* block-device below us           */

	/*
	** bitmap administration to register which blocks are modified
	*/
	int		mapunit; 	/* blocksize per bit in bitmap     */
	long int	mapsize;	/* total size of bitmap (bytes)    */
	long int	mapremain;	/* remaining bytes in last bitmap  */
	int		mapcount;       /* number of bitmaps in use        */
	char 		*mapcache[MAPMAXNR];	/* pointers to bitmaps     */

	char		*iobuf;		/* buffer of <mapunit> bytes       */

	/*
	** administration for interface with the kernel-thread
	*/
	int		pid;		/* pid==0: no thread available    */
	int		closedown;	/* boolean: thread exit required  */
	short		qfilled;	/* boolean: I/O request pending   */
	struct request	*req;		/* request to be handled now      */
	int		iobusy;		/* boolean: rq under treatment    */
	wait_queue_head_t waitq;	/* wait-Q: thread waits for work  */

	/*
	** statistical counters
	*/
	unsigned long	nrreads;	/* number of  read-actions        */
	unsigned long	nrwrites;	/* number of write-actions        */
	unsigned long	nrcowblocks;	/* number of blocks in use on cow */
};

static struct cowloop_device	cowlodev;

/*
** function to be called by core-kernel to handle the I/O-requests
** in the queue
*/
static void
cowlo_request(request_queue_t *q)
{
#ifdef	COWDEBUG
	printk("cowloop - request function called....\n");
#endif

	while (1)
	{
		if(cowlodev.iobusy)
			return; 

#ifdef	COWDEBUG
		printk("cowloop - get next request\n");
#endif
	 	/*
		** get next request from queue;
		** this macro issues a return if request-queue is empty
		*/
		INIT_REQUEST;

		cowlodev.iobusy	= 1;

		/*
		** valid request ?
		*/
		if (	MINOR(CURRENT->rq_dev) != 0		||
       		 	CURRENT->rq_status     == RQ_INACTIVE	||
			CURRENT->sector        == -1  		  )
		{
#ifdef	COWDEBUG
			printk("cowloop - request failed (not valid)\n");
#endif
			end_request(0);		/* request failed */
			cowlodev.iobusy	= 0;
			continue;
		}


		/*
		** drop the system-wide request-lock which has been
		** acquired by the kernel before calling the request
		** function
		*/
		spin_unlock_irq(&io_request_lock);

		/*
		** when no kernel-thread is available, the request will
		** produce an I/O-error
		*/
		if (!cowlodev.pid)
		{
			printk(KERN_ERR "cowloop - no thread available\n");
			end_request(0);		/* request failed */
			cowlodev.iobusy	= 0;
			continue;
		}

		/*
		** handle I/O-request in the context of the kernel-thread
		*/
		cowlodev.req 		= CURRENT;
		cowlodev.qfilled	= 1;

		// wake_up_interruptible_sync(&cowlodev.waitq);
		wake_up_interruptible(&cowlodev.waitq);

		spin_lock_irq(&io_request_lock);

		/*
		** get out of this function now while the I/O-request is
		** under treatment of the kernel-thread; this function
		** will be called again after the current I/O-request has
		** been finished by the thread
		*/
		return;
	}
}

/*
** function to be called in the context of the thread to handle
** the queued I/O-requests 
**
** return-value:	0 - fail, 1 - success
*/
static long int
do_request(struct request *req)
{
	unsigned long	len, mapnum, bytenum, bitnum, blocknr,
			partlen, mu = cowlodev.mapunit;
	long int	rv, l;
	char		*p, *mc;
	loff_t 		offset, pos;

	/*
	** if our blksize_size has been modified by upper kernel-layers
	** (e.g. VFS), also modify the blksize_size of the driver below us
	*/
	if (cowlodev.belowdev)
	{
		if (cowloop_blocksz != *cowlodev.undersz)
		{
#ifdef	COWDEBUG
			printk(KERN_NOTICE
			       "cowloop - change blocksize %d -> %d\n",
			       	*cowlodev.undersz, cowloop_blocksz);
#endif
			set_blocksize(cowlodev.belowdev, cowloop_blocksz);
		}
	}

	/*
	** calculate some variables which are needed later on
	*/
	len     =        req->current_nr_sectors << 9;
	offset  = (pos_t)req->sector             << 9;

#ifdef	COWDEBUG
	printk("cowloop - req cmd=%d offset=%lld len=%lu addr=%p\n",
				req->cmd, offset, len, req->buffer);
#endif

	/*
	** handle READ- or WRITE-request
	*/
	switch (req->cmd) 
	{
	   /**********************************************************/
	   case READ:
		cowlodev.nrreads++;

		switch ( cowlo_checkio(len, offset) )
		{
		   case ALLCOW:
#ifdef	COWDEBUG
			printk("cowloop - read straight allcow\n");
#endif
			rv = cowlo_readcow(req->buffer, len, offset);
			break;

		   case ALLRDO:
#ifdef	COWDEBUG
			printk("cowloop - read straight allrdo\n");
#endif
			rv = cowlo_readrdo(req->buffer, len, offset);
			break;

	   	   case MIXEDUP:
			/*
			** somewhat more complicated approach: breakup required
			*/
			l	= len;
   			p	= req->buffer;
			pos	= offset;

			for (rv=1; l > 0; l-=partlen, p+=partlen, pos+=partlen)
			{
				/*
				** calculate blocknr of entire block
				*/
				blocknr = (pos_t)pos / mu;
	
				/*
				** calculate partial length for this transfer
				*/
				partlen = mu - (pos_t)pos % mu;
				if (partlen > l)
					partlen = l;
	
				/*
				** is this block located in the cow-file
				*/
				mapnum  = CALCMAP (blocknr);
				bytenum = CALCBYTE(blocknr);
				bitnum  = CALCBIT (blocknr);
				mc	= cowlodev.mapcache[mapnum];
	
				if (*(mc+bytenum)&(1<<bitnum))
				{
					/*
					** read (partial) block from cow-file
					*/
#ifdef	COWDEBUG
					printk("cowloop - split readcow "
					       "partlen=%ld off=%lld\n",
					       partlen, pos);
#endif
					if (cowlo_readcow(p, partlen, pos) <= 0)
						rv = 0;
				}
				else
				{
					/*
					** read (partial) block from rdo-file
					*/
#ifdef	COWDEBUG
					printk("cowloop - split readrdo "
					       "partlen=%ld off=%lld\n",
					       partlen, pos);
#endif
					if (cowlo_readrdo(p, partlen, pos) <= 0)
						rv = 0;
				}
			}
			break;

		   default:
			rv = 0;	/* never happens */
		}
		break;

	   /**********************************************************/
	   case WRITE:
		cowlodev.nrwrites++;

		switch ( cowlo_checkio(len, offset) )
		{
		   case ALLCOW:
			/*
			** straight-forward write will do...
			*/
#ifdef	COWDEBUG
			printk("cowloop - write straight allcow\n");
#endif
			rv = cowlo_writecow(req->buffer, len, offset);
			break;	/* from switch */

		   case ALLRDO:
			if (len%mu == 0)
			{
#ifdef	COWDEBUG
				printk("cowloop - write straight allrdo\n");
#endif
				rv = cowlo_writecow(req->buffer, len, offset);
				break;
			}
	   	   case MIXEDUP:
			/*
			** somewhat more complicated stuff is required:
			** if the request is larger than one underlying
			** block or is spread over two underlying blocks,
			** split the request into pieces; if a block doesnot
			** start at a block boundary, take care that 
			** surrounding data is read first (if needed),
			** fit the new data in and write it as a full block
			*/
			l	= len;
   			p	= req->buffer;
			pos	= offset;

			for (rv=1; l > 0; l-=partlen, p+=partlen, pos+=partlen)
			{
				/*
				** calculate partial length for this transfer
				*/
				partlen = mu - (pos_t)pos % mu;
				if (partlen > l)
					partlen = l;

				/*
				** calculate blocknr of entire block
				*/
				blocknr = (pos_t)pos / mu;

				/*
				** has this block been written before?
				*/
				mapnum  = CALCMAP (blocknr);
				bytenum = CALCBYTE(blocknr);
				bitnum  = CALCBIT (blocknr);
				mc	= cowlodev.mapcache[mapnum];

				if (*(mc+bytenum)&(1<<bitnum))
				{
					/*
					** block has been written before;
					** write transparantly to cow file
					*/
#ifdef	COWDEBUG
					printk("cowloop - split write transp\n");
#endif
					if (cowlo_writecow(p, partlen, pos)<=0)
						rv = 0;
				}
				else
				{
					/*
					** block has never been written before,
					** so read entire block from
					** read-only file first, unless
					** a full block is requested to
					** be written
					*/
					if (partlen < mu)
					{
						if (cowlo_readrdo(
							cowlodev.iobuf,
							mu, blocknr*mu) <= 0)
							rv = 0;
					}
	
					/*
					** transfer modified part into
					** the block just read
					*/
					memcpy(cowlodev.iobuf + (pos_t)pos%mu,
								p, partlen);

					/*
					** write entire block to cow file
					*/
#ifdef	COWDEBUG
					printk("cowloop - split partlen=%ld "
					       "off=%lu\n",
						partlen, (pos_t)blocknr*mu);
#endif
					if (cowlo_writecow(cowlodev.iobuf,
					           mu, (pos_t)blocknr*mu) <= 0)
						rv = 0;
				}
			}
			break;

		   default:
			rv = 0;	/* never happens */
		}
		break;

	   default:
		printk(KERN_ERR
		       "cowloop - unrecognized command %d\n", req->cmd);
		rv = 0;
	}

	return (rv <= 0 ? 0 : 1);
}

/*
** read data from the read-only file
**
** return-value: similar to user-mode read
*/
static long int
cowlo_readrdo(char *buf, int len, loff_t pos)
{
	long int	rv;
	mm_segment_t	old_fs;
	loff_t		savepos = pos;

#ifdef	COWDEBUG
	printk("cowloop - call readrdo %p\n", cowlodev.rdofp->f_op->read);
#endif
        old_fs = get_fs();
	set_fs( get_ds() );
	rv = cowlodev.rdofp->f_op->read(cowlodev.rdofp, buf, len, &pos);
        set_fs(old_fs);

	if (rv < len)
		printk(KERN_WARNING "cowloop - read-failure %ld on rdofile"
		                    "- offset=%lld len=%d\n",
					rv, savepos, len);
	return rv;
}

/*
** read data from the cow file with
** a modified offset (skipping the bitmap)
**
** return-value: similar to user-mode read
*/
static long int
cowlo_readcow(char *buf, int len, loff_t pos)
{
	long int	rv;
	mm_segment_t	old_fs;
	loff_t		tmpoffset;

#ifdef	COWDEBUG
	printk("cowloop - call readcow %p\n", cowlodev.cowfp->f_op->read);
#endif
	tmpoffset = pos + cowlodev.mapsize;

        old_fs = get_fs();
	set_fs( get_ds() );
	rv = cowlodev.cowfp->f_op->read(cowlodev.cowfp, buf, len, &tmpoffset);
        set_fs(old_fs);

	if (rv < len)
		printk(KERN_WARNING "cowloop - read-failure %ld on cowfile"
		                    "- offset=%lld len=%d\n",
					rv, pos+cowlodev.mapsize, len);
	return rv;
}

/*
** write data to the cow file with a modified offset (skipping the bitmap);
** if the concerning block(s) is/are written for the first time,
** update the bitmap in memory and flush it to the cow file
**
** return-value: similar to user-mode write
*/
static long int
cowlo_writecow(char *buf, int len, loff_t pos)
{
	long int		rv, rv2, blockstart, bitsset;
	unsigned long int	mapnum=0, bytenum=0, bitnum=0, blocknr=0,
				partlen, mu = cowlodev.mapunit;
	loff_t			tmpoffset;
	mm_segment_t		old_fs;

	/*
	** write the entire block to the cow file
	*/
#ifdef	COWDEBUG
	printk("cowloop - call writecow %p\n", cowlodev.cowfp->f_op->write);
#endif
	tmpoffset = pos + cowlodev.mapsize;

        old_fs = get_fs();
	set_fs( get_ds() );
	rv = cowlodev.cowfp->f_op->write(cowlodev.cowfp, buf, len, &tmpoffset);
        set_fs(old_fs);

	if (rv < len)
		printk(KERN_WARNING "cowloop - write-failure %ld on cowfile"
		                    "- offset=%lld len=%d\n",
					rv, pos+cowlodev.mapsize, len);
	if (rv <= 0)
		return rv;

#ifdef	COWDEBUG
	printk("cowloop - block written\n");
#endif

	/*
	** check if underlying block(s) is/are written in the cow file
	** for the first time; if so, adapt the bitmap
	*/
	for (bitsset=0; len > 0; len-=partlen, pos+=partlen)
	{
		/*
		** calculate partial length for this transfer
		*/
		partlen = mu - (pos_t)pos % mu;
		if (partlen > len)
			partlen = len;
	
		/*
		** calculate blocknr of underlying block
		*/
		blocknr = (pos_t)pos / mu;

		mapnum  = CALCMAP (blocknr);
		bytenum = CALCBYTE(blocknr);
		bitnum  = CALCBIT (blocknr);

		if (*(cowlodev.mapcache[mapnum]+bytenum) & (1<<bitnum))
			continue;	/* already written before */

	       	/*
		** if the block is written for the first time,
		** the corresponding bit should be set in the bitmap
		** and that bitmap block should be written again
		*/
		*(cowlodev.mapcache[mapnum]+bytenum) |= (1<<bitnum);

		bitsset++;
		cowlodev.nrcowblocks++;

#ifdef	COWDEBUG
		printk("cowloop - bitupdate blk=%ld map=%ld byte=%ld bit=%ld\n",
					blocknr, mapnum, bytenum, bitnum);
#endif
	}

	/*
	** if any bit is modified in the bitmap, that bitmap-block should
	** be flushed to disk
	**
	** notice that in a rare situation bits could have been set in
	** different bitmap-blocks (or even in different in different
	** maps); in that case not all bitmaps are flushed (bitmaps will
	** be flushed with a next update or anyhow when unloading the
	** cowloop-driver)
	*/
	if (bitsset)
	{
		blockstart = bytenum / cowlodev.mapunit * cowlodev.mapunit;
		tmpoffset  = mapnum * MAPMAXSZ + blockstart;

       		old_fs = get_fs();
		set_fs( get_ds() );
		rv2 = cowlodev.cowfp->f_op->write(cowlodev.cowfp,
				cowlodev.mapcache[mapnum]+blockstart,
				cowlodev.mapunit, &tmpoffset);
		set_fs(old_fs);

		if (rv2 < 0)
		{
			printk(KERN_WARNING
				"cowloop - write-failure %ld on bitmap - "
				"blk=%ld map=%ld byte=%ld bit=%ld\n",
				rv2, blocknr, mapnum, bytenum, bitnum);
		}
#ifdef	COWDEBUG
		printk("cowloop - bitmap written\n");
#endif
	}

	return rv;
}

/*
** check for a given I/O-request if all underlying blocks 
** (with size <mapunit>) are either in the read-only file or in
** the cow file
*/
static int
cowlo_checkio(int len, loff_t offset)
{
	unsigned long	mapnum, bytenum, bitnum, blocknr,
			partlen, mu = cowlodev.mapunit;
	long int	totcnt, cowcnt, l;
	char		*mc;
	loff_t 		pos;

	/*
	** notice that the requested block might cross
	** a blocksize boundary while one of the concerned
	** blocks resides in the read-only file and another
	** one in the copy-on-write file; in that case the
        ** request will be broken up into pieces
	*/
	if (len <= cowlodev.mapunit		  	&&
	    cowlodev.mapunit - (pos_t)offset % cowlodev.mapunit <= len)
	{
		/*
		** easy situation:
		** requested data-block entirely fits within
		** the mapunit used for the bitmap
		** check if that block is located in rdo-file or
		** cow-file
		*/
		blocknr = (pos_t)offset/cowlodev.mapunit;

		mapnum  = CALCMAP (blocknr);
		bytenum = CALCBYTE(blocknr);
		bitnum  = CALCBIT (blocknr);

		if (*(cowlodev.mapcache[mapnum]+bytenum)&(1<<bitnum))
			return ALLCOW;
		else
			return ALLRDO;
	}

	/*
	** less easy situation:
	** the requested data-block does not fit within the mapunit
	** used for the bitmap
	** check if *all* underlying blocks involved reside on the rdo-file
       	** or the cow-file (so still no breakup required)
	*/
	l	= len;
	pos	= offset;

	for (cowcnt=0, totcnt=0; l > 0; l-=partlen, pos+=partlen, totcnt++)
	{
		/*
		** calculate blocknr of involved block
		*/
		blocknr = (pos_t)pos / mu;

		/*
		** calculate partial length for this transfer
		*/
		partlen = mu - ((pos_t)pos % mu);
		if (partlen > l)
			partlen = l;

		/*
		** is this block located in the cow-file
		*/
		mapnum  = CALCMAP (blocknr);
		bytenum = CALCBYTE(blocknr);
		bitnum  = CALCBIT (blocknr);

		mc	= cowlodev.mapcache[mapnum];

		if (*(mc+bytenum)&(1<<bitnum))
			cowcnt++;;

#ifdef	COWDEBUG
		printk("cowloop - check %lu - map %lu, byte %lu, bit %lu, "
		       "cowcnt %ld, totcnt %ld %02x %p\n",
			blocknr, mapnum, bytenum, bitnum, cowcnt, totcnt,
			*(mc+bytenum), mc);
#endif
	}

	if (cowcnt == 0)	/* all involved blocks on rdo-file? */
		return ALLRDO;

	if (cowcnt == totcnt)	/* all involved blocks on cow-file? */
		return ALLCOW;

	/*
	** situation somewhat more complicated:
	** involved underlying blocks spread over both files
	*/
	return MIXEDUP;
}

/*
** daemon-process (kernel-thread) executes this function
*/
static int
cowlo_daemon(void *vp)
{
	int	rv;

        strcpy(current->comm, "cowloopd");
        daemonize();

	while (!cowlodev.closedown)
	{
		/*
		** sleep while waiting for an I/O request;
		** note that no non-interruptible wait has been used
		** because the non-interruptible version of
		** a *synchronous* wake_up does not exist (any more)
		*/
		if (wait_event_interruptible(cowlodev.waitq, cowlodev.qfilled))
		{
			flush_signals(current);
			continue;
		}

		if (cowlodev.closedown)		/* module will be unloaded ? */
		{
			cowlodev.pid = 0;
			return 0;
		}

		/*
		** woken up by the I/O-request handler:	treat requested I/O
		*/
		cowlodev.qfilled = 0;
	
		rv = do_request(cowlodev.req);
	
		/*
		** reacquire the generic spin-lock for manipulating
		** the request-queue and dequeue the request
		*/
		spin_lock_irq(&io_request_lock);
	
		end_request(rv);
		cowlodev.iobusy = 0;

		/*
		** if there are more requests in the queue,
		** initiate the next one
		*/
		if (!QUEUE_EMPTY)
			cowlo_request(NULL);

		spin_unlock_irq(&io_request_lock);
	}

	return 0;
}

/*
** readproc-function: called when the corresponding /proc-file is read
*/
static int
cowlo_readproc(char *buf, char **start, off_t pos, int cnt, int *eof, void *p)
{
	return sprintf(buf,
		"    read-only file: %9s\n"
		"copy-on-write file: %9s\n"
		"     read-requests: %9ld\n"
		"    write-requests: %9ld\n"
		"     bitmap-blocks: %9ld (of %d bytes)\n"
		" cow-blocks in use: %9ld (of %d bytes)\n"
		" cowloop-blocksize: %9d bytes\n"
		"     pid of thread: %9d\n",
			rdofile,
			cowfile,
			cowlodev.nrreads,
			cowlodev.nrwrites,
			cowlodev.mapsize/cowlodev.mapunit, cowlodev.mapunit,
			cowlodev.nrcowblocks, cowlodev.mapunit,
			cowloop_blocksz,
			cowlodev.pid);
}

/*
** handle system call open()/mount()
*/
static int
cowlo_open(struct inode *inode, struct file *file)
{
	if (!inode)
		return -EINVAL;

	if (MAJOR(inode->i_rdev) != COWMAJOR)
	{
		printk(KERN_WARNING "cowlo_open: pseudo-major != %d\n",
		     						  COWMAJOR);
		return -ENODEV;
	}

	if (MINOR(inode->i_rdev) != 0)
		return -ENODEV;

	MOD_INC_USE_COUNT;

#ifdef	COWDEBUG
	printk("cowloop - open\n");
#endif

	return 0;
}

/*
** handle system call close()/umount()
*/
static int
cowlo_release(struct inode *inode, struct file *file)
{
	int err = 0;

	if (!inode)
		return 0;

	if (MINOR(inode->i_rdev) == 0)
  		err = fsync_dev(inode->i_rdev);

	MOD_DEC_USE_COUNT;

#ifdef	COWDEBUG
	printk("cowloop - release\n");
#endif
	return err;
}

/*
** called by init_module to open files and allocate bitmaps
*/
static int
init_cowloop(void)
{
	struct file	*f;
	struct inode	*inode;
	long int	nb, i, rv;
	loff_t		offset;
	mm_segment_t	old_fs;

	/*
	** open the read-only file
	** 'rdofile' is module option string
	*/
        if(rdofile == NULL) 
        {
     	     	printk(KERN_ERR "cowloop - specify name of read-only file "
		                "as argument 'rdofile=filename'\n\n");
         	return 0;
        }

	f = filp_open(rdofile, O_RDONLY, 0);

	if (f == NULL || IS_ERR(f))
	{
		printk(KERN_ERR
		       "cowloop - unable to get file %s for fs\n", rdofile);
		return 0;
	}

	inode = f->f_dentry->d_inode;

	if (!S_ISREG(inode->i_mode) && !S_ISBLK(inode->i_mode))
	{
		printk(KERN_ERR "cowloop - %s not regular file or blockdev\n",
							rdofile);
		filp_close(f,0);
		return 0;
	}

	cowlodev.rdofp = f;

	/*
	** determine block-size and total size of read-only file
	*/
	cowloop_blocksz = 0;

	if (S_ISREG(inode->i_mode))
	{	
		/*
		** read-only file: regular file
		*/
		kdev_t	d = inode->i_dev;

		cowlodev.belowdev	= 0;

		cowlodev.numblocks = inode->i_size >> BLOCK_SIZE_BITS;

		if (blksize_size[MAJOR(d)])
			cowloop_blocksz = blksize_size[MAJOR(d)][MINOR(d)];

		if (cowloop_blocksz == 0)
			cowloop_blocksz = BLOCK_SIZE;
	}
	else
	{	/*
		** read-only file: block device
		*/
		kdev_t	d = inode->i_rdev;

		if (blksize_size[MAJOR(d)])
		{
			cowloop_blocksz   = blksize_size[MAJOR(d)][MINOR(d)];

			cowlodev.belowdev = d;
			cowlodev.undersz  = blksize_size[MAJOR(d)]+MINOR(d);
		}
		else
		{
			cowlodev.belowdev = 0;
		}

		if (blk_size[MAJOR(d)])
			cowlodev.numblocks = blk_size[MAJOR(d)][MINOR(d)];
		else
			cowlodev.numblocks = 0;
	}

	/*
	** open copy-on-write file (read-write)
	** 'cowfile' is module option string
	*/
        if(cowfile == NULL)
        {
          	printk(KERN_ERR
                 "cowloop - specify name of copy-on-write file "
		 "as argument 'cowfile=filename'\n\n");

	  	filp_close(cowlodev.rdofp, 0);
         	return 0;
        }

	f = filp_open(cowfile, O_RDWR, 0600);

	if (f == NULL || IS_ERR(f))
	{
		/*
		** non-existing cowfile: try to create
		*/
		f = filp_open(cowfile, O_RDWR|O_CREAT, 0600);

		if (f == NULL || IS_ERR(f))
		{
			printk(KERN_ERR
		       	  "cowloop - failed to open file %s for read-write\n\n",
		       						cowfile);
		  	filp_close(cowlodev.rdofp, 0);
			return 0;
	 	}
	}

	inode = f->f_dentry->d_inode;

	if (!S_ISREG(inode->i_mode))
	{
		printk(KERN_ERR "cowloop - %s is not regular file\n", cowfile);
		filp_close(f, 0);
	  	filp_close(cowlodev.rdofp, 0);
		return 0;
	}

	cowlodev.cowfp = f;

	/*
	** cow file layout:
	** 	+-----------------------------+
	**	|     used-block bitmap       |
	**	|                             |
	**	|-----------------------------|  <---- start-offset cow blocks
	**      |       written blocks        |
	**      |          .....              |
	**
	** used-block bitmap:
	**   - contains one bit per block of size <cowlodev.mapunit>
	**   - bit-value '1' = block has been written on cow
	**               '0' = block unused on cow
	**   - total bitmap rounded to multiples of <cowlodev.mapunit>
	**
	** calculate size (in bytes) for total bitmap in cow file
	** (i.e. the start-offset for the data blocks)
	*/
	cowlodev.mapunit = 1024;

	nb = ((unsigned long int)cowlodev.numblocks << BLOCK_SIZE_BITS)
						/ cowlodev.mapunit;

	if (nb%8)	/* transform #bits to #bytes */
		nb+=8;  /* rounded if necessary      */
	nb /= 8;

	if (nb%cowlodev.mapunit)	/* round up #bytes to blocks */
		cowlodev.mapsize = (nb/cowlodev.mapunit+1) * cowlodev.mapunit;
	else
		cowlodev.mapsize = nb;

	/*
	** be sure that a new cow file gets the right size by writing
	** the last byte of the file (unless the file has already the
	** right size); the cow file should have the size of the
	** input file plus the size of the bitmap
	*/
	offset = ((pos_t)cowlodev.numblocks << BLOCK_SIZE_BITS)
						+ cowlodev.mapsize - 1;

	if (inode->i_size != 0 && inode->i_size != offset+1)
	{
		printk(KERN_ERR
		       "cowloop - size of cowfile should be 0 or %lld bytes\n",
								offset+1);
		printk(KERN_ERR
		       "cowloop - cowfile %s (size %lld) not related "
		       "to rdofile %s\n", cowfile, inode->i_size, rdofile);
		  	filp_close(cowlodev.rdofp, 0);
		  	filp_close(cowlodev.cowfp, 0);
			return 0;
	}

	if (inode->i_size == 0)
	{
        	old_fs = get_fs();
		set_fs( get_ds() );
		rv=cowlodev.cowfp->f_op->write(cowlodev.cowfp, "", 1, &offset);
		set_fs(old_fs);

		if (rv < 1)
		{
			printk(KERN_ERR
				"cowloop - cannot set cow-file to size %lld\n",
								offset+1);
		  	filp_close(cowlodev.rdofp, 0);
		  	filp_close(cowlodev.cowfp, 0);
			return 0;
		}
	}

	/*
	** reserve space in memory as transfer-buffer
	*/
	cowlodev.iobuf  = kmalloc(cowlodev.mapunit, GFP_KERNEL);

	if (!cowlodev.iobuf)
	{
		printk(KERN_ERR "cowloop - cannot get space for buffer %d\n",
							cowlodev.mapunit);
	  	filp_close(cowlodev.rdofp, 0);
	  	filp_close(cowlodev.cowfp, 0);
		return 0;
	}

	/*
	** reserve space in memory for the entire bitmap and
	** fill it with the bitmap-data from disk; the entire
	** bitmap is allocated in several chunks because kmalloc
	** has restrictions regarding the allowed size per kmalloc
	*/
	cowlodev.mapcount = (cowlodev.mapsize+MAPMAXSZ-1)/MAPMAXSZ;

	if (cowlodev.mapcount > MAPMAXNR)
	{
		printk(KERN_ERR "cowloop - not enough bitmaps available "
				"totmapsz=%ld, mapcnt=%d mapunit=%d\n",
				cowlodev.mapsize, cowlodev.mapcount,
				cowlodev.mapunit);

		kfree(cowlodev.iobuf);
	  	filp_close(cowlodev.rdofp, 0);
	  	filp_close(cowlodev.cowfp, 0);
		return 0;
	}

	/*
	** the size of every bitmap chunk will be MAPMAXSZ bytes, except for
	** the last bitmap chunk: calculate remaining size for this chunk
	*/
	if (cowlodev.mapsize % MAPMAXSZ == 0)
		cowlodev.mapremain = MAPMAXSZ;
	else
		cowlodev.mapremain = cowlodev.mapsize % MAPMAXSZ;

	/*
	** allocate space for all bitmap chunks in memory
	*/
	for (i=0; i < cowlodev.mapcount; i++)
	{
		if (i < (cowlodev.mapcount-1))
			cowlodev.mapcache[i]  = kmalloc(MAPMAXSZ, GFP_KERNEL);
		else
			cowlodev.mapcache[i]  = kmalloc(cowlodev.mapremain,
						                  GFP_KERNEL);

		if (!cowlodev.mapcache[i])
		{
			printk(KERN_ERR "cowloop - no space for bitmap "
					"totmapsz=%ld, mapcnt=%d mapunit=%d\n",
					cowlodev.mapsize, cowlodev.mapcount,
					cowlodev.mapunit);

			for (i--; i >= 0; i--)		/* undo mallocs */
				kfree(cowlodev.mapcache[i]);

			kfree(cowlodev.iobuf);

		  	filp_close(cowlodev.rdofp, 0);
		  	filp_close(cowlodev.cowfp, 0);
			return 0;
		}
	}

	/*
	** read the entire in-memory bitmap from the cow file;
	** count the number of blocks filled already (statistical purposes)
	*/
	cowlodev.nrcowblocks = 0;

        old_fs = get_fs();
	set_fs( get_ds() );

	for (i=0, offset=0; i < cowlodev.mapcount; i++)
	{
		long	numbytes;
		char	*p;

		if (i < (cowlodev.mapcount-1))
			/*
			** full bitmap chunk
			*/
			numbytes = MAPMAXSZ;
		else
			/*
			** last bitmap chunk: might be partly filled
			*/
			numbytes = cowlodev.mapremain;

		cowlodev.cowfp->f_op->read(cowlodev.cowfp,
				cowlodev.mapcache[i], numbytes, &offset);

		/*
		** count the bits which have been set for statistical
		** purposes
		*/
		p = cowlodev.mapcache[i];

		for (numbytes--; numbytes >= 0; numbytes--, p++)
		{
			/*
			** for only eight checks the following construction
			** is faster than a loop-construction
			*/
			if ((*p) & 0x01)	cowlodev.nrcowblocks++;
			if ((*p) & 0x02)	cowlodev.nrcowblocks++;
			if ((*p) & 0x04)	cowlodev.nrcowblocks++;
			if ((*p) & 0x08)	cowlodev.nrcowblocks++;
			if ((*p) & 0x10)	cowlodev.nrcowblocks++;
			if ((*p) & 0x20)	cowlodev.nrcowblocks++;
			if ((*p) & 0x40)	cowlodev.nrcowblocks++;
			if ((*p) & 0x80)	cowlodev.nrcowblocks++;
		}
	}

	set_fs(old_fs);

	return 1;
}

static struct block_device_operations cowlo_fops =
{
        .open    =     cowlo_open,	/* called upon open  */
        .release =     cowlo_release,	/* called upon close */
};

/*
** called during insmod
*/
int
init_module(void)
{
	if (!init_cowloop())
		return -EINVAL;

	if (register_blkdev(COWMAJOR, DEVICE_NAME, &cowlo_fops))
	{
		printk(KERN_WARNING
			"cowloop - unable to get major %d for cowloop\n",
		       						COWMAJOR);

                if (cowlodev.rdofp)
                {
                	filp_close(cowlodev.rdofp, 0);
                	cowlodev.rdofp = NULL;
                }

                if (cowlodev.cowfp)
                {
                	filp_close(cowlodev.cowfp, 0);
                	cowlodev.cowfp = NULL;
                }

		return -EIO;
	}

	blk_init_queue(BLK_DEFAULT_QUEUE(COWMAJOR), DEVICE_REQUEST);

	init_waitqueue_head(&cowlodev.waitq);

	cowloop_maxsect	= cowloop_blocksz / cowloop_hardsect;

	blksize_size[COWMAJOR] = &cowloop_blocksz;
	blk_size[COWMAJOR]     = &(cowlodev.numblocks);
	hardsect_size[COWMAJOR]= &cowloop_hardsect;
	max_sectors[COWMAJOR]  = &cowloop_maxsect;
	read_ahead[COWMAJOR]   =  cowloop_maxsect;

	/*
	** create a /proc-entry to produce status-info on request
	*/
	create_proc_read_entry("cowloop", 0, 0, cowlo_readproc, NULL);

	cowlodev.pid = kernel_thread(cowlo_daemon, 0, 0);

	/*
	** report cowloop status
	*/
	revision[sizeof revision - 3] = '\0';

	printk(KERN_NOTICE "cowloop - (C) 2004 AT Consultancy/AT Computing\n");
	printk(KERN_NOTICE "cowloop - version %s\n", &revision[11]);
	printk(KERN_NOTICE "cowloop - info: www.ATConsultancy.nl/cowloop\n");
	printk(KERN_NOTICE "cowloop -      read-only file: %9s\n", rdofile);
	printk(KERN_NOTICE "cowloop - size read-only file: %9d Kb\n",
		       					*blk_size[COWMAJOR]);
	printk(KERN_NOTICE "cowloop -  copy-on-write file: %9s (%s)\n",
			cowfile, cowlodev.nrcowblocks ? "existing" : "new");
	printk(KERN_NOTICE "cowloop -   cow blocks in use: %9ld Kb\n",
			   cowlodev.nrcowblocks * (cowlodev.mapunit/1024));
	printk(KERN_NOTICE "cowloop -          bitmapsize: %9ld Kb (%d maps)\n",
			   	cowlodev.mapsize/1024, cowlodev.mapcount);
	return 0;
}

/*
** called during rmmod
*/
void
cleanup_module(void) 
{
	int		i;
	mm_segment_t	old_fs;
	loff_t		offset;

	/*
	** flush the entire bitmap to the cow-file;
	** bitmap-chunks containing all zeroes are not flushed
	*/
        old_fs = get_fs();
	set_fs( get_ds() );

	for (i=0, offset=0; i < cowlodev.mapcount; i++)
	{
		long	numbytes, j;

		if (i < (cowlodev.mapcount-1))
			/*
			** full bitmap chunk
			*/
			numbytes = MAPMAXSZ;
		else
			/*
			** last bitmap chunk: might be partly filled
			*/
			numbytes = cowlodev.mapremain;

		/*
		** check if entire bitmap-chunk contains binary zeroes
		*/
		for (j=0; j < numbytes; j++)
			if (*(cowlodev.mapcache[i]+j))
				break;

		if (j==numbytes)
		{
			offset += numbytes;
			continue;	/* skip writing: hole remains hole */
		}

		printk(KERN_NOTICE "cowloop - flushing bitmap %2d (%3ld Kb)\n",
						i, numbytes/1024);

		if (cowlodev.cowfp->f_op->write(cowlodev.cowfp,
			cowlodev.mapcache[i], numbytes, &offset) < numbytes)
			break;
	}
	set_fs(old_fs);

	/*
	** wakeup kernel-thread to be able to exit
	** and wait until it has exited
	*/
	cowlodev.closedown = 1;
	cowlodev.qfilled   = 1;
	wake_up_interruptible(&cowlodev.waitq);

        while (cowlodev.pid)
                schedule();

	/*
	** get rid of /proc-entry and unregister the driver
	*/
	remove_proc_entry("cowloop", NULL);

	if (unregister_blkdev(COWMAJOR, DEVICE_NAME) != 0)
		printk(KERN_WARNING "cowloop - cannot unregister blkdev\n");

	blk_cleanup_queue(BLK_DEFAULT_QUEUE(COWMAJOR));

	/*
	** close the read-only file and cow file
	*/
	if (cowlodev.rdofp)
		filp_close(cowlodev.rdofp, 0);

	if (cowlodev.cowfp)
		filp_close(cowlodev.cowfp, 0);

	blk_size[COWMAJOR]     = 0;
	blksize_size[COWMAJOR] = 0;

	/*
	** release all kmalloc'ed memory
	*/
	for (i=0; i < cowlodev.mapcount; i++)
		kfree(cowlodev.mapcache[i]);

	kfree(cowlodev.iobuf);

	printk(KERN_NOTICE "cowloop - unloaded\n");
}
